require 'dir2svn/version'
require 'find'
require 'fileutils'

module Dir2svn
  ERR1 = 'Provide the source and target path before doing a sync'
  ERR2 = 'Source folder does not exist'
  ERR3 = 'Target needs to be in a Subversion managed folder'

  # Utility to sync a folder to a Subversion working copy
  # The working copy folder is the 'target'
  class Sync
    attr_accessor :source, :target, :exclude, :do_puts

    def initialize(*args)
      @source = args[0]
      @target = args[1]
      @exclude = args[2] || '.svn,.trashbin,.folder,__generated.txt'
      if Gem.win_platform?
        @source = args[0].tr('\\', '/')
        @target = args[1].tr('\\', '/')
        # Ruby style paths work on MS Windows
        # when interacting directory with OS using system calls
      end
      @do_puts = args[3] || true
    end

    def check_source_target_etc
      fail ERR1 if source.nil? || target.nil?
      unless File.exist? source
        puts "Source: #{source}" if do_puts
        fail ERR2
      end
    end

    def svn_mkdir_target
      `svn mkdir --parents \"#{target}\" 2>>svn_mkdir.tmp 1>>svn_mkdir.tmp`
      svn_mkdir = File.read('svn_mkdir.tmp')
      File.delete('svn_mkdir.tmp')
      fail ERR3 if svn_mkdir.include? 'is not a working copy'
    end

    def run
      check_source_target_etc
      svn_mkdir_target
      results = Dir2svn.sync_paths(source, target, exclude, do_puts)
    end
  end

  def sync_paths(src, dest, exclude, do_puts)
    src = File.expand_path(src) + '/'
    dest = File.expand_path(dest) + '/'
    results = nil
    # First files, then folders,
    # as the processing of files can also sync folders
    results = Dir2svn.sync_file_paths(src, dest, exclude, do_puts)
    Dir2svn.sync_folder_paths(src, dest, exclude, do_puts)
    results
  end
  module_function :sync_paths

  # See SvnRepository#sync_paths
  def sync_file_paths(src, dest, exclude, do_puts)
    fls_dest = Dir2svn.find_files(dest, exclude, do_puts).collect { |p| p.sub(dest, '') }
    fls_src = Dir2svn.find_files(src, exclude, do_puts).collect { |p| p.sub(src, '') }
    fls_d = fls_dest - fls_src # deleted: in DEST but not (anymore) in SRC
    fls_a = fls_src - fls_dest # added: in SRC but not (yet) in SRC
    fls_f = fls_src - fls_a # found: in SRC and DEST: everything in SRC that we didn't add
    fls_u = [] # updated
    fls_e = [] # equal

    # Update files
    fls_f.collect do |p| # update files
      p1 = File.expand_path(p, src) # new version
      p0 = File.expand_path(p, dest) # existing version
      if File.identical?(p1, p0)
        fls_e << p
      else
        FileUtils.cp(p1, p0)
        fls_u << p
      end
    end

    # Add files
    fls_a.each do |p|
      p1 = File.expand_path(p, src) # new version
      p0 = File.expand_path(p, dest)
      unless File.exist?(File.dirname(p0)) # create parent dir of new file unless it exists
        # self.makedir_parents(File.dirname(p0)) unless File.exists?(File.dirname(p0))
        cmd = "svn mkdir --parents \"#{File.dirname(p0)}\""
        out = `#{cmd}`
        puts "#{cmd}\n#{out}" if do_puts
      end
      FileUtils.cp(p1, p0)
      cmd = "svn add \"#{p0}\""
      out = `#{cmd}`
      puts "#{cmd}\n#{out}" if do_puts
    end

    # delete files
    fls_d.each do |p|
      p0 = File.expand_path(p, dest)
      cmd = "svn rm \"#{p0}\""
      out = `#{cmd}`
      puts "#{cmd}\n#{out}" if do_puts
    end
    { deleted: fls_d, added: fls_a, updated: fls_u, equal: fls_e }
  end
  module_function :sync_file_paths

  # See SvnRepository#sync_paths
  def sync_folder_paths(src, dest, exclude, do_puts)
    fldrs_src = Dir2svn.find_folders(src, exclude, do_puts).collect { |p| p.sub(src, '') }
    fldrs_dest = Dir2svn.find_folders(dest, exclude, do_puts).collect { |p| p.sub(dest, '') }
    fldrs_d = (fldrs_dest - fldrs_src) # deleted folders: see deleted files
    fldrs_a = (fldrs_src - fldrs_dest) # add folders: see added files
    fldrs_d.each do |p| # deleted folders
      p0 = File.expand_path(p, dest)
      next unless File.exist?(p0)
      cmd = "svn rm \"#{p0}\""
      out = `#{cmd}`
      puts "#{cmd}\n#{out}" if do_puts
    end
    fldrs_a.each do |p| # added folders (must be empty, not very useful but..)
      p0 = File.expand_path(p, dest)
      cmd = "svn mkdir --parents \"#{p0}\""
      out = `#{cmd}`
      puts "#{cmd}\n#{out}" if do_puts
    end
  end
  module_function :sync_folder_paths

  def find_files(path, exclude, do_puts)
    Dir2svn.find_paths(path, exclude, do_puts)[:files]
  end
  module_function :find_files

  def find_folders(path, exclude, do_puts)
    Dir2svn.find_paths(path, exclude, do_puts)[:folders]
  end
  module_function :find_folders

  def find_paths(path, exclude, _do_puts)
    fls = []
    fldrs = []
    Find.find(path) do |p|
      if FileTest.directory?(p)
        if exclude.include?(File.basename(p))
          Find.prune # Don't look any further into this directory.
        else
          fldrs << p
          next
        end
      else # ignore .folder files
        fls << p unless exclude.include?(File.basename(p))
      end
    end
    { files: fls, folders: fldrs }
  end
  module_function :find_paths
end
